/**
	@file	sc_query_parser.rl
	@author Ted Schmidt
	@date   10-08-2007

	Sharpcast Query Language state machine gneration code
	Copyright Sharpcast, Inc. 2002-2007. All rights reserved.
 */

package com.sharpcast.datastore.common;

import java.util.Vector;
import java.util.Hashtable;

public class QueryParser {
	// Test op names
	public static final String SC_AND_OP_NAME = "ScAndOp.ScTestOp";
	public static final String SC_OR_OP_NAME = "ScOrOp.ScTestOp";
	public static final String SC_EQUAL_OP_NAME = "ScEqualOp.ScTestOp";
	public static final String SC_NOT_EQUAL_OP_NAME = "ScNotEqualOp.ScTestOp";
	public static final String SC_GREATER_THAN_OP_NAME = "ScGreaterThanOp.ScTestOp";
	public static final String SC_GREATER_THAN_EQUAL_OP_NAME = "ScGreaterThanEqualOp.ScTestOp";
	public static final String SC_LESS_THAN_OP_NAME = "ScLessThanOp.ScTestOp";
	public static final String SC_LESS_THAN_EQUAL_OP_NAME = "ScLessThanEqualOp.ScTestOp";
	public static final String SC_CASELESS_COMPARE_OP_NAME = "ScCaselessCompareOp.ScTestOp";
	public static final String SC_FIELDS_NOT_EQUAL_OP_NAME = "ScFieldsNotEqualOp.ScTestOp";
	public static final String SC_SUB_OBJECT_OP_NAME = "ScSubObjectOp.ScTestOp";
	public static final String SC_MAP_OP_NAME = "ScMapOp.ScTestOp";

	// Membership spec names
	public static final String SC_CHILD_SPEC_NAME = "ScChildSpec.ScMembershipSpec";
	public static final String SC_RECURSIVE_CHILD_SPEC_NAME = "ScRecursiveChildSpec.ScMembershipSpec";
	public static final String SC_PARENT_SPEC_NAME = "ScParentSpec.ScMembershipSpec";

	// Modifier names
	public static final String SC_RESFIRST_QUERY_NAME = "start-index";
	public static final String SC_RESFIRST_RECORD_NAME = "resfirst";
	public static final String SC_RESMAX_QUERY_NAME = "max-results";
	public static final String SC_RESMAX_RECORD_NAME = "resmax";
	public static final String SC_ORDERBY_QUERY_NAME = "orderby";
	public static final String SC_ORDERBY_RECORD_NAME = "order_by_fld";

	// Modifier defaults
	public static final int SC_RESFIRST_DEFAULT = 0;
	public static final int SC_RESMAX_DEFAULT = -1;

	// OpTypes
	public static final int SC_OTYPE_NULL = 0;
	public static final int SC_OTYPE_AND = 1;
	public static final int SC_OTYPE_OR = 2;
	public static final int SC_OTYPE_EQ = 3;
	public static final int SC_OTYPE_CASELESSCOMP = 4;
	public static final int SC_OTYPE_GTEQ = 5;
	public static final int SC_OTYPE_GT = 6;
	public static final int SC_OTYPE_LTEQ = 7;
	public static final int SC_OTYPE_LT = 8;
	public static final int SC_OTYPE_NEQ = 9;
	public static final int SC_OTYPE_SUBOBJ = 10;
	public static final int SC_OTYPE_MAP = 11;
	public static final int SC_OTYPE_FNEQ = 12;

	// MapTypes
	public static final int SC_MAP_TYPE_NULL = 0;
	public static final int SC_MAP_TYPE_SINGLE = 1;
	public static final int SC_MAP_TYPE_ANY = 2;
	public static final int SC_MAP_TYPE_INDIRECT = 3;

	// QtreeTypes
	public static final int SC_QTREE_TYPE_EMPTY = 0;
	public static final int SC_QTREE_TYPE_OPENPAREN = 1;
	public static final int SC_QTREE_TYPE_CLOSEPAREN = 2;
	public static final int SC_QTREE_TYPE_OR = 3;
	public static final int SC_QTREE_TYPE_OP = 4;

	private class ScQueryTreeToken {
		Record m_pOp;
		int m_iType;
	}

	private class NullVariant {
		// Placeholder object for nulls
	}

	private String m_strQuery;
	private int m_pchVarBegin;
	private int m_pchVarEnd;
	private boolean m_bVar;
	private long m_iVar;
	private int m_iVarType;
	private boolean m_fIVarIsNeg;
	private boolean m_fHasDefault;
	private boolean m_bIsQuery;

	private int m_iOpType;
	private int m_iSubOpType;
	private int m_iMapOpType;

	private Vector m_vecVars;
	private Hashtable m_mapModifiers;

	private Vector m_vecTrees;

	private Record m_Op;
	private Record m_mSpec;

	private boolean m_fHasMSpec;
	private boolean m_fHasTestOp;

	private int m_iLine;
	private int m_iCol;
	private int m_cs;

	private void reset() {
		m_strQuery 		= null;
		m_pchVarBegin	= 0;
		m_pchVarEnd 	= 0;
		m_bVar 			= false;
		m_iVar 			= 0;
		m_iVarType 		= Record.VALUE_TYPE_INVALID;
		m_fIVarIsNeg 	= false;
		m_fHasDefault 	= false;
		m_bIsQuery 		= false;

		m_iOpType 		= SC_OTYPE_NULL;
		m_iSubOpType 	= SC_OTYPE_NULL;
		m_iMapOpType 	= SC_MAP_TYPE_NULL;

		m_vecVars 		= new Vector();
		m_mapModifiers 	= new Hashtable();
		m_vecTrees 		= new Vector();

		m_Op 			= null;
		m_mSpec 		= null;

		m_fHasMSpec 	= false;
		m_fHasTestOp 	= false;

		m_iLine 		= 0;
		m_iCol 			= 0;
		m_cs 			= 0;
	}

	public Record createQuery (String[] pstrTestSpecs, int nTestSpecs,
			UnsignedLong maskNotify, String szPath) throws DatastoreException
	{
		Vector vecRecs = new Vector();
		Record rec = null;
		Record recOut = new Record();
		int i = 0;

		for (i = 0; i < nTestSpecs; i++) {
			reset();
			rec = createTestSpec(pstrTestSpecs[i]);

			if (rec != null) {
				vecRecs.addElement(rec);
				rec = null;
			}
		}

		recOut.setType("ScQuery");
		recOut.setValue("chid_min", new Long (0));
		recOut.setValue("id", new UnsignedLong (0));
		recOut.setValue("ignore_chid_min", new Boolean (false));
		recOut.setValue("notify_mask", maskNotify);
		recOut.setValue("test_specs", vecRecs);
		recOut.setValue("path", new Path (szPath));

		return recOut;
	}

	private Record createTestSpec (String szQuery) throws DatastoreException
	{
		Object	pvar = null;
		Vector 	vecRoot = null;
		Record	recOut = null;

		vecRoot = new Vector ();
		m_vecTrees.addElement (vecRoot);

		m_strQuery = szQuery;
		parse (szQuery);

		if (!m_bIsQuery)
			return null;

		recOut = new Record ();
		
		recOut.setType("ScQuery::TestSpec");

		// Any subtrees should be built by this point, there should
		// only be one tree on the stack
		if (m_vecTrees.size() != 1)
			error ("invalid query syntax");
		m_Op = treeBuild ();

		if (m_fHasTestOp)
			recOut.setValue ("testop", m_Op);
		else
			recOut.setNullValue ("testop");

		if (m_fHasMSpec)
			recOut.setValue ("member_spec", m_mSpec);
		else
			recOut.setNullValue ("member_spec");

		// Add modifiers

		pvar = m_mapModifiers.get(SC_RESFIRST_QUERY_NAME);
		if (pvar != null)
			recOut.setValue (SC_RESFIRST_RECORD_NAME, pvar);
		else
			recOut.setValue (SC_RESFIRST_RECORD_NAME, new Long (SC_RESFIRST_DEFAULT));

		pvar = m_mapModifiers.get(SC_RESMAX_QUERY_NAME);
		if (pvar != null)
			recOut.setValue (SC_RESMAX_RECORD_NAME, pvar);
		else
			recOut.setValue (SC_RESMAX_RECORD_NAME, new Long (SC_RESMAX_DEFAULT));
		
		pvar = m_mapModifiers.get(SC_ORDERBY_QUERY_NAME);
		if (pvar != null)
			recOut.setValue (SC_ORDERBY_RECORD_NAME, pvar);
		else
			recOut.setNullValue (SC_ORDERBY_RECORD_NAME);
		
		// Type should be the last thing on the stack at this point
		if (m_vecVars.size () != 1)
			error ("incorrect number of tokens");
		pvar = m_vecVars.lastElement();
		recOut.setValue ("type", pvar);

		return recOut;
	}

	public void addSubtree() {
		m_vecTrees.addElement(new Vector());
	}

	public void error (String strError) throws DatastoreException
	{
		throw new DatastoreException (strError);
	}

	public void pushVar() throws DatastoreException
	{

		switch (m_iVarType) 
		{
		case Record.VALUE_TYPE_UINT:
			m_vecVars.addElement(new UnsignedLong(m_iVar));
			break;
		case Record.VALUE_TYPE_INT:
			m_vecVars.addElement(new Long(m_iVar));
			break;
		case Record.VALUE_TYPE_STRING:
			m_vecVars.addElement(new String(m_strQuery.substring(m_pchVarBegin,
					m_pchVarEnd)));
			break;
		case Record.VALUE_TYPE_PATH:
			m_vecVars.addElement(new Path(m_strQuery.substring (m_pchVarBegin, m_pchVarEnd)));
			break;
		case Record.VALUE_TYPE_BOOL:
			m_vecVars.addElement(new Boolean(m_bVar));
			break;
		case Record.VALUE_TYPE_NULL:
			m_vecVars.addElement(new NullVariant());
			break;
		default:
			error("Invalid variant type");
		}

	}

	public void pushOpType(int iOpType) 
	{
		if (m_iOpType == SC_OTYPE_SUBOBJ)
			m_iSubOpType = iOpType;
		else
			m_iOpType = iOpType;
	}

	public Record createOp () throws DatastoreException 
	{
		Record recOut = null;
		
		switch (m_iOpType)
		{
		case SC_OTYPE_EQ:
			recOut = pushCompOp (SC_EQUAL_OP_NAME);
			break;
		case SC_OTYPE_CASELESSCOMP:
			recOut = pushCompOp (SC_CASELESS_COMPARE_OP_NAME);
			break;
		case SC_OTYPE_GTEQ:
			recOut = pushCompOp (SC_GREATER_THAN_EQUAL_OP_NAME);
			break;
		case SC_OTYPE_GT:
			recOut = pushCompOp (SC_GREATER_THAN_OP_NAME);
			break;
		case SC_OTYPE_LTEQ:
			recOut = pushCompOp (SC_LESS_THAN_EQUAL_OP_NAME);
			break;
		case SC_OTYPE_LT:
			recOut = pushCompOp (SC_LESS_THAN_OP_NAME);
			break;
		case SC_OTYPE_NEQ:
			recOut = pushCompOp (SC_NOT_EQUAL_OP_NAME);
			break;
		case SC_OTYPE_SUBOBJ:
			recOut = pushSubObjOp ();
			break;
		case SC_OTYPE_MAP:
			recOut = pushMapOp (m_iMapOpType);
			break;
		case SC_OTYPE_FNEQ:
			recOut = pushFneqOp ();
			break;
		default:
			error ("Invalid op type");
		}

		// Reset our hasDefault flag for the next op (unless it's a subobj)
		if (m_iOpType != SC_OTYPE_SUBOBJ)
			m_fHasDefault = false;
		// Reset our optype
		m_iOpType = SC_OTYPE_NULL;
		return recOut;
	}

	private Record	pushCompOp (String szOpName) throws DatastoreException 
	{
		Object pvar = null;
		Record recOut = new Record ();

		recOut.setType(szOpName);

		setDefault (recOut);
		
		pvar = m_vecVars.lastElement();
		m_vecVars.removeElementAt(m_vecVars.size() - 1);
		if (pvar instanceof NullVariant)
			recOut.setNullValue("value");
		else
			recOut.setValue("value", pvar);
			
		
		pvar = m_vecVars.lastElement();
		m_vecVars.removeElementAt(m_vecVars.size() - 1);
		if (pvar instanceof NullVariant)
			recOut.setNullValue("fname");
		else
			recOut.setValue("fname", pvar);
			
		
		return recOut;
	}

	private Record pushMapOp (int type) throws DatastoreException
	{
		Object	pvar = null;
		Record	recOut = new Record();
		Record	recInner = null;

		recOut.setType(SC_MAP_OP_NAME);

		// Create the SubOp(s)
		if (m_vecTrees.size() <= 1)
			error ("query syntax error");
		recInner = treeBuild ();
		
		recOut.setValue("map_entry_op", recInner);

		switch (type)
		{
		case SC_MAP_TYPE_SINGLE:
			pvar = m_vecVars.lastElement();
			m_vecVars.removeElementAt(m_vecVars.size() - 1);
			recOut.setValue("map_key_name", pvar);
			recOut.setValue("mode", new UnsignedLong (SC_MAP_TYPE_SINGLE));
			recOut.setValue("map_key_field_name", "");
			break;
		case SC_MAP_TYPE_INDIRECT:
			pvar = m_vecVars.lastElement();
			m_vecVars.removeElementAt(m_vecVars.size() - 1);
			recOut.setValue("map_key_field_name", pvar);
			recOut.setValue("mode", new UnsignedLong (SC_MAP_TYPE_INDIRECT));
			recOut.setValue("map_key_name", "");
			break;
		case SC_MAP_TYPE_ANY:
			recOut.setValue("mode", new UnsignedLong (SC_MAP_TYPE_ANY));
			recOut.setValue("map_key_name", "");
			recOut.setValue("map_key_field_name", "");
			break;
		default:
			error ("bad map type");
		}
		pvar = m_vecVars.lastElement();
		m_vecVars.removeElementAt(m_vecVars.size() - 1);
		recOut.setValue ("map_field_name", pvar);

		m_iMapOpType = SC_MAP_TYPE_NULL;
		return recOut;
	}

	private Record pushSubObjOp() throws DatastoreException 
	{
		Object pvar = null;
		Record recInner = null;
		Record recOut = new Record();

		recOut.setType(SC_SUB_OBJECT_OP_NAME);

		// Create the SubOp
		m_iOpType = m_iSubOpType;
		recInner = createOp();

		recOut.setValue("inner_op", recInner);

		pvar = m_vecVars.lastElement();
		m_vecVars.removeElementAt(m_vecVars.size() - 1);
		recOut.setValue("field_name", pvar);

		return recOut;
	}

	private Record pushFneqOp() throws DatastoreException 
	{
		Object pvar = null;
		Record recOut = new Record();

		recOut.setType(SC_FIELDS_NOT_EQUAL_OP_NAME);

		setDefault(recOut);

		pvar = m_vecVars.lastElement();
		m_vecVars.removeElementAt(m_vecVars.size() - 1);
		recOut.setValue("field_right", pvar);
		pvar = m_vecVars.lastElement();
		m_vecVars.removeElementAt(m_vecVars.size() - 1);
		recOut.setValue("field_left", pvar);

		return recOut;
	}

	private void setDefault(Record rec) throws DatastoreException
	{
		if (m_fHasDefault) {
			Object pvar = m_vecVars.lastElement();
			m_vecVars.removeElementAt(m_vecVars.size() - 1);
			rec.setValue("default", pvar);
		} else {
			rec.setValue("default", new Boolean (false));
		}
	}

	private void pushChildSpec () throws DatastoreException
	{
		Object pvar = null;
		
		if (m_fHasMSpec) // We can't have more than one
			error ("Multiple member specs defined");

		m_fHasMSpec = true;
		m_mSpec = new Record ();

		m_mSpec.setType (SC_CHILD_SPEC_NAME);

		m_mSpec.setValue ("recursive", new Boolean (false));

		pvar = m_vecVars.lastElement();
		m_vecVars.removeElementAt(m_vecVars.size() - 1);
		m_mSpec.setValue ("vec_name", pvar);
		
		pvar = m_vecVars.lastElement();
		m_vecVars.removeElementAt(m_vecVars.size() - 1);
		m_mSpec.setValue("vec_id", pvar);
	}

	private void pushParentSpec() throws DatastoreException
	{
		Object pvar = null;

		if (m_fHasMSpec) // We can't have more than one
			error("Multiple member specs defined");

		m_fHasMSpec = true;
		m_mSpec = new Record ();

		m_mSpec.setType(SC_PARENT_SPEC_NAME);

		m_mSpec.setValue("recursive", new Boolean (false));
		pvar = m_vecVars.lastElement();
		m_vecVars.removeElementAt(m_vecVars.size() - 1);
		m_mSpec.setValue("vec_name", pvar);
		m_mSpec.setValue("vec_id", new Path(""));
	}

	public void charDigit(int iDigit, int iBase) {
		if (m_fIVarIsNeg)
			m_iVar = m_iVar * iBase - iDigit;
		else
			m_iVar = m_iVar * iBase + iDigit;
	}

	public void addModifier() {
		Object pvarKey = m_vecVars.lastElement();
		m_vecVars.removeElementAt(m_vecVars.size() - 1);
		Object pvarVal = m_vecVars.lastElement();
		m_vecVars.removeElementAt(m_vecVars.size() - 1);

		m_mapModifiers.put(pvarKey, pvarVal);
	}

	public void treePushOpenParen() 
	{
		ScQueryTreeToken token = new ScQueryTreeToken();
		token.m_iType = SC_QTREE_TYPE_OPENPAREN;
		((Vector) m_vecTrees.lastElement()).addElement(token);
	}

	public void treePushCloseParen() 
	{
		ScQueryTreeToken token = new ScQueryTreeToken();
		token.m_iType = SC_QTREE_TYPE_CLOSEPAREN;
		((Vector) m_vecTrees.lastElement()).addElement(token);
	}

	public void treePushOr() 
	{
		ScQueryTreeToken token = new ScQueryTreeToken();
		token.m_iType = SC_QTREE_TYPE_OR;
		((Vector) m_vecTrees.lastElement()).addElement(token);
	}

	public void treePushOp() throws DatastoreException 
	{
		ScQueryTreeToken token = new ScQueryTreeToken();
		token.m_iType = SC_QTREE_TYPE_OP;
		token.m_pOp = createOp();
		((Vector) m_vecTrees.lastElement()).addElement(token);
	}

	private Record treeBuild () throws DatastoreException
	{
		Vector tree = (Vector)m_vecTrees.lastElement();
		m_vecTrees.removeElementAt(m_vecTrees.size() - 1);
		while (tree.size() > 1)
			subTreeBuild (tree, 0, tree.size() - 1);
		
		if (tree.size() == 0) {
			// there is no test op.
			return null;
		} else {
			m_fHasTestOp = true;
			return ((ScQueryTreeToken)tree.elementAt(0)).m_pOp;
		}
	}
	
	private int subTreeBuild (Vector tree, int nStartToken, int nEndToken) throws DatastoreException
	{
		int nRemovedInner = 0;
		int nRemovedOuter = 0;
		for (int nCur = nStartToken; nCur <= nEndToken; nCur++) {
			switch (((ScQueryTreeToken) tree.elementAt(nCur)).m_iType) {
			case SC_QTREE_TYPE_OPENPAREN:
				nRemovedInner = handleOpenParen(tree, nCur, nEndToken);
				break;
			case SC_QTREE_TYPE_CLOSEPAREN:
				// skip it
				break;
			case SC_QTREE_TYPE_OR:
				if (nCur == nStartToken)
					error("| character in invalid location");
				break;
			case SC_QTREE_TYPE_OP:
				nRemovedInner = handleOp (tree, nCur, nEndToken);
				break;
			default:
				error("parser error");
			}
			nRemovedOuter += nRemovedInner;
			nEndToken -= nRemovedInner;
			nRemovedInner = 0;		
		}
		return nRemovedOuter;
	}
	
	private int handleOpenParen (Vector tree, int nStartToken, int nEndToken) throws DatastoreException
	{
		int nCur;
		int nRemovedInner = 0;
		int nRemovedOuter = 0;
		for (nCur = nStartToken; nCur <= nEndToken; nCur++)
		{
			if (((ScQueryTreeToken)tree.elementAt(nCur)).m_iType == SC_QTREE_TYPE_CLOSEPAREN)
			{
				// if there's just one op in the parens, strip the parens
				if (nCur == nStartToken + 2)
				{
					tree.removeElementAt(nCur);
					tree.removeElementAt(nStartToken);
					nRemovedInner = 2;
				}
				else // subtree build the stuff in the parens
				{
					nRemovedInner = subTreeBuild (tree, nStartToken + 1, nCur - 1);
				}
				nRemovedOuter += nRemovedInner;
				nEndToken -= nRemovedInner;
				nRemovedInner = 0;
				break;
			}
		}
		if (nCur > nEndToken + nRemovedOuter)
			error ("unmatched parentheses, extra \"(\"");
		return nRemovedOuter;
	}
	
	private int handleOp (Vector tree, int nStartToken, int nEndToken) throws DatastoreException
	{
		int nRemoved = 0;
		if (nEndToken > nStartToken)
		{
			Record boolOp = null;
			switch (((ScQueryTreeToken) tree.elementAt(nStartToken+1)).m_iType) {
			case SC_QTREE_TYPE_OP: // AND the two together
				boolOp = createBoolOp(true,
						((ScQueryTreeToken) tree.elementAt(nStartToken)).m_pOp,
						((ScQueryTreeToken) tree.elementAt(nStartToken + 1)).m_pOp);
				((ScQueryTreeToken) tree.elementAt(nStartToken)).m_pOp = boolOp;
				tree.removeElementAt(nStartToken + 1);
				nRemoved = 1;
				break;
			case SC_QTREE_TYPE_OR: // OR the two together as long as precedence permits
				if (nEndToken > nStartToken + 2)
				{
					if (((ScQueryTreeToken) tree.elementAt(nStartToken + 2)).m_iType == SC_QTREE_TYPE_OP
							&& ((ScQueryTreeToken) tree.elementAt(nStartToken + 3)).m_iType != SC_QTREE_TYPE_OP) {
						boolOp = createBoolOp(false,
							((ScQueryTreeToken) tree.elementAt(nStartToken)).m_pOp,
							((ScQueryTreeToken) tree.elementAt(nStartToken + 2)).m_pOp);
					((ScQueryTreeToken) tree.elementAt(nStartToken)).m_pOp = boolOp;
					tree.removeElementAt(nStartToken + 2);
					tree.removeElementAt(nStartToken + 1);
					nRemoved = 2;
					}
				}
				else if (nEndToken > nStartToken + 1)
				{
					if (((ScQueryTreeToken) tree.elementAt(nStartToken + 2)).m_iType == SC_QTREE_TYPE_OP) {
						boolOp = createBoolOp(false,
							((ScQueryTreeToken) tree.elementAt(nStartToken)).m_pOp,
							((ScQueryTreeToken) tree.elementAt(nStartToken + 2)).m_pOp);
					((ScQueryTreeToken) tree.elementAt(nStartToken)).m_pOp = boolOp;
					tree.removeElementAt(nStartToken + 2);
					tree.removeElementAt(nStartToken + 1);
					nRemoved = 2;
					}
					else
					{
						error ("| character in incorrect location");
					}
				}
				else
				{
					error ("| character in incorrect location");
				}
				break;
			case SC_QTREE_TYPE_OPENPAREN:
			case SC_QTREE_TYPE_CLOSEPAREN:
				// ignore, get it next time around
				break;
				default:
					error ("parser error");
			}
		}
		return nRemoved;
	}

	private Record createBoolOp (boolean fIsAnd, Record opLeft, Record opRight) throws DatastoreException
	{
		Record recOut = new Record ();
		if (fIsAnd)
			recOut.setType (SC_AND_OP_NAME);
		else // OR
			recOut.setType (SC_OR_OP_NAME);
		
		recOut.setValue ("op_right", opLeft);
		recOut.setValue	("op_left", opRight);
		return recOut;
	}


%%{
	machine sc_query_parser;
	
	alphtype char;
	variable cs m_cs;
	variable p pch;
	variable pe pchEnd;
	variable data szQuery;

	
	action Err
	{
		error ("Syntax Error");
	}
	
	action LineCount
	{
		if (szQuery[pch] == '\n')
		{
			m_iLine++;
			m_iCol = 1;
		}
		else
		{
			m_iCol++;
		}
	}
	
	action SetTypeUint
	{
		m_iVarType = Record.VALUE_TYPE_UINT;
	}
	
	action SetTypeInt
	{
		m_iVarType = Record.VALUE_TYPE_INT;
	}
	
	action SetTypeStr
	{
		m_iVarType = Record.VALUE_TYPE_STRING;
	}
	
	action SetTypePath
	{
		m_iVarType = Record.VALUE_TYPE_PATH;
	}
	
	action SetTypeBool
	{
		m_iVarType = Record.VALUE_TYPE_BOOL;
	}
	
	action VarBegin
	{
		m_pchVarBegin = 0;
		m_pchVarEnd = 0;
		m_iVar = 0;
		m_fIVarIsNeg = false;
		m_bVar = false;
		m_iVarType = Record.VALUE_TYPE_NULL;
	}
	
	action VarComplete
	{
		pushVar ();
	}
	
	action VarNull
	{
		pushNullVar ();
	}
	
	action OpEq
	{
		pushOpType (SC_OTYPE_EQ);
	}
	
	action OpCaselessComp
	{
		pushOpType (SC_OTYPE_CASELESSCOMP);
	}
	
	action OpGtEq
	{
		pushOpType (SC_OTYPE_GTEQ);
	}
	
	action OpGt
	{
		pushOpType (SC_OTYPE_GT);
	}
	
	action OpLtEq
	{
		pushOpType (SC_OTYPE_LTEQ);
	}
	
	action OpLt
	{
		pushOpType (SC_OTYPE_LT);
	}
	
	action OpNEq
	{
		pushOpType (SC_OTYPE_NEQ);
	}
	
	action OpSubObj
	{
		pushOpType (SC_OTYPE_SUBOBJ);
	}
	
	action OpMap
	{
		pushOpType (SC_OTYPE_MAP);
	}
	
	action OpFNEq
	{
		pushOpType (SC_OTYPE_FNEQ);
	}
	
	action ChildSpecComplete
	{
		pushChildSpec ();
	}

	action ParentSpecComplete
	{
		pushParentSpec ();
	}
	
	action AddSubTree
	{
		addSubtree ();
	}
	
	action SetOpNull
	{
		m_fHasTestOp = false;
	}
	
	action DecimalCharDigit
	{
		charDigit (szQuery[pch] - '0', 10);
	}

	action HexCharDigit
	{
		charDigit (szQuery[pch] - '0', 16);
	}
	
	action HexCharLower
	{
		charDigit (10 + szQuery[pch] - 'a', 16);
	}
	
	action HexCharUpper
	{
		charDigit (10 + szQuery[pch] - 'A', 16);
	}

	action SetVarTrue
	{
		m_bVar = true;
	}
	
	action SetVarFalse
	{
		m_bVar = false;
	}
	
	action StrBegin
	{
		m_pchVarBegin = fpc;
	}
	
	action StrEnd
	{
		m_pchVarEnd = fpc;
	}
	
	action StrEndEOF
	{
		m_pchVarEnd = pchEnd;
	}
	
	action SetNegativeFlag
	{
		m_fIVarIsNeg = true;
	}
	
	action SetDefault
	{
		m_fHasDefault = true;
	}
	
	action TreePushOp
	{
		treePushOp ();
	}
	
	action TreePushOr
	{
		treePushOr ();
	}
	
	action TreePushOpenParen
	{
		treePushOpenParen ();
	}
	
	action TreePushCloseParen
	{
		treePushCloseParen ();
	}
	
	action ModifierComplete
	{
		addModifier ();
	}
	
	action IsQuery
	{
		m_bIsQuery = true;
	}
	
	action SetMapOpTypeSingle
	{
		m_iMapOpType = SC_MAP_TYPE_SINGLE;
	}
	
	action SetMapOpTypeAny
	{
		m_iMapOpType = SC_MAP_TYPE_ANY;
	}
	
	action SetMapOpTypeIndirect
	{
		m_iMapOpType = SC_MAP_TYPE_INDIRECT;
	}

	HighChar = (0x80..0xFFFD);

	Char = 0x9 | 0xA | 0xD | (0x20..0x7F) | HighChar;

	S = (0x20 | 0x9 | 0xD | 0xA)+;

	Sign = '-' %SetNegativeFlag;

	DecimalNumber = Sign? (digit+) @DecimalCharDigit;
	
	HexNumber = '0x'(digit @HexCharDigit | [a-f] @HexCharLower | [A-F] @HexCharUpper)+ %SetTypeUint %/SetTypeUint;
	
	String = "'" (Char-['\\] | "\\"Char)* >StrBegin %StrEnd "'";
	
	StringOrPath = (String			%SetTypeStr %/SetTypeStr
					| String 'p'	%SetTypePath %/SetTypePath
					); 
					
	DecimalIntOrUint = (DecimalNumber		%SetTypeInt %/SetTypeInt
						| DecimalNumber 'u'	%SetTypeUint %/SetTypeInt
						);
	
	Null = 'NULL'i;
	
	TrueFalse = ('true'		%SetVarTrue %/SetVarTrue
				| 'false'	%SetVarFalse %/SetVarFalse
				)			%SetTypeBool %/SetTypeBool
				;

	Var = (StringOrPath | DecimalIntOrUint | HexNumber | TrueFalse | Null) >VarBegin	%VarComplete %/VarComplete;
	
	Ident =  (((alpha | '_') (alnum | '_' | '.' | '-')*) - ('true' | 'false' | 'NULL')) >VarBegin >StrBegin %StrEnd %SetTypeStr %VarComplete %/StrEndEOF %/SetTypeStr %/VarComplete;
	
	CompOperator =	( '=='	%OpEq
					| ':'	%OpEq
					| '~='	%OpCaselessComp
					| '>='	%OpGtEq
					| '>'	%OpGt
					| '<='	%OpLtEq
					| '<'	%OpLt
					| '!='	%OpNEq
					);
					
	Default = ('(' S? TrueFalse S? ')') >VarBegin %SetDefault %VarComplete;

	MapSpec =	( Ident %SetMapOpTypeIndirect
				| Var	%SetMapOpTypeSingle
				| '*'	%SetMapOpTypeAny
				);

	FieldOrSubField =	( Ident
						| Ident '#' Ident %OpSubObj
						);
	
	Op = '\[' S? FieldOrSubField S? CompOperator S? (Var | Ident %OpFNEq) S? Default? S? '\]' @TreePushOp;
	
	MapOp = '\[' S? Ident '{' MapSpec '}' S? '#' S? %AddSubTree (Op S?)+ %OpMap '\]' @TreePushOp;
	
	ChildSpec = '{' S? 'childof:'i S? Var S? ',' S? Ident S? '}' @ChildSpecComplete;
	
	ParentSpec = '{' S? 'haschildren:'i S? Ident S? '}' @ParentSpecComplete;
	
	MSpec = (ChildSpec | ParentSpec);
	
	Modifier = '&' Ident '=' (Var | Ident) %ModifierComplete %/ModifierComplete;
	
	OpOrRelation =	( Op
					| MapOp	
					| '|'	%TreePushOr
					| '('	%TreePushOpenParen
					| ')'	%TreePushCloseParen %/TreePushCloseParen
					);
	
	Query = S? (Ident S? (MSpec S?)? (OpOrRelation S?)* (Modifier S?)*) >IsQuery;
	
	main := Query >!Err $LineCount;

}%%

%% write data;	

	private boolean parse (String pchQuery) throws DatastoreException
	{
		int pch = 0;
		int pchEnd = pch + pchQuery.length ();
		char[] szQuery = pchQuery.toCharArray ();
		
		%%{
			write init;
			write exec;
			write eof;
		}%%
		
		return (pch == pchEnd);
	}
}